import os
import sys
import logging
import click
from flask_compress import Compress
from klue.swagger.apipool import ApiPool
from klue_microservice.log import set_level
from klue_microservice.api import do_ping
from klue_microservice.crash import set_error_reporter, crash_handler
from klue_microservice.exceptions import format_error
from klue_microservice.auth import set_jwt_defaults


log = logging.getLogger(__name__)


#
# API: class to define then run a micro service api
#

class API(object):


    def __init__(self, app, host='localhost', port=80, debug=False, log_level=logging.DEBUG, formats=None, timeout=20, error_reporter=None, jwt_secret=None, jwt_audience=None):
        """Take the flask app, and optionally the http port to listen on, and
        whether flask's debug mode is one or not, which callback to call when
        catching exceptions, and the api's log level"""
        assert app
        self.app = app
        self.port = port
        self.host = host
        self.debug = debug
        self.formats = formats
        self.timeout = timeout
        set_jwt_defaults(secret=jwt_secret, audience=jwt_audience)
        set_level(log_level)
        if error_reporter:
            set_error_reporter(error_reporter)
        log.info("Initialized API (%s:%s) (Flask debug:%s)" % (host, port, debug))


    def load_apis(self, path, ignore=[]):
        """Load all swagger files found at the given path, except those whose
        names are in the 'ignore' list"""

        if not path:
            raise Exception("Missing path to api swagger files")

        if type(ignore) is not list:
            raise Exception("'ignore' should be a list of api names")

        # Find all swagger apis under 'path'
        apis = {}

        log.debug("Searching path %s" % path)
        for root, dirs, files in os.walk(path):
            for f in files:
                if f.endswith('.yaml'):
                    api_name = f.replace('.yaml', '')

                    if api_name in ignore:
                        log.info("Ignoring api %s" % api_name)
                        continue

                    apis[api_name] = os.path.join(path, f)
                    log.debug("Found api %s in %s" % (api_name, f))

        # And add klue-microservice's default ping api
        ping_path = os.path.join(os.path.dirname(sys.modules[__name__].__file__), 'ping.yaml')
        apis['ping'] = ping_path

        # Save found apis
        self.path_apis = path
        self.apis = apis


    def start(self, serve=[]):
        """Start the API server"""

        # Check arguments
        if type(serve) is str:
            serve = [serve]
        elif type(serve) is list:
            pass
        else:
            raise Exception("'serve' should be an api name or a list of api names")

        if len(serve) == 0:
            raise Exception("You must specify at least one api to serve")

        for api_name in serve:
            if api_name not in self.apis:
                raise Exception("Can't find %s.yaml (swagger file) in the api directory %s" % (api_name, self.path_apis))

        app = self.app
        app.secret_key = os.urandom(24)

        # Always serve the ping api
        serve.append('ping')

        # Let's compress returned data when possible
        compress = Compress()
        compress.init_app(app)

        # All apis that are not served locally are not persistent
        not_persistent = []
        for api_name in self.apis.keys():
            if api_name in serve:
                pass
            else:
                not_persistent.append(api_name)

        # Now load those apis into the ApiPool
        for api_name, api_path in self.apis.items():

            host = None
            port = None

            if api_name in serve:
                # We are serving this api locally: override the host:port specified in the swagger spec
                host = self.host
                port = self.port

            do_persist = True if api_name not in not_persistent else False
            local = True if api_name in serve else False

            log.info("Loading api %s from %s (persist: %s)" % (api_name, api_path, do_persist))
            ApiPool.add(
                api_name,
                yaml_path=api_path,
                timeout=self.timeout,
                error_callback=format_error,
                formats=self.formats,
                do_persist=do_persist,
                host=host,
                port=port,
                local=local,
            )

        ApiPool.merge()

        # Now spawn flask routes for all endpoints
        for api_name in self.apis.keys():
            if api_name in serve:
                log.info("Spawning api %s" % api_name)
                api = getattr(ApiPool, api_name)
                api.spawn_api(app, decorator=crash_handler)

        if os.path.basename(sys.argv[0]) == 'gunicorn':
            # Gunicorn takes care of spawning workers
            return

        # Debug mode is the default when not running via gunicorn
        app.debug = self.debug
        app.run(host='0.0.0.0', port=port)

#
# Generic code to start server, from command line or via gunicorn
#

def letsgo(name, callback=None):
    assert callback

    @click.command()
    @click.option('--port', help="Set server listening port (default: 80)", default=80)
    @click.option('--debug/--no-debug', default=True)
    def main(port, debug):
        callback(port, debug)

    if name == "__main__":
        main()

    if os.path.basename(sys.argv[0]) == 'gunicorn':
        callback()



import logging
import json
import uuid
import sys
import traceback
import inspect
from functools import wraps
from pprint import pformat
from flask import request, Response
from klue.swagger.apipool import ApiPool
from klue_microservice.utils import timenow
from klue_microservice.exceptions import UnhandledServerError


log = logging.getLogger(__name__)


try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack


#
# Default error reporting
#

def default_error_reporter(title, message):
    """By default, error messages are just logged"""
    log.error("error: %s" % title)
    log.error("details:\n%s" % message)

error_reporter = default_error_reporter

def set_error_reporter(callback=None):
    """Here you can override the default crash reporter (and send yourself
    emails, sms, slacks...)"""

    if error_reporter:
        global error_reporter
        error_reporter = callback


def report_error(data, msg=None, caught=None, title=None):
    """Format a crash report and send it somewhere relevant. There are two
    types of crashes: fatal crashes (backend errors) or non-fatal ones (just
    reporting a glitch, but the api call did not fail)"""

    if isinstance(data, dict):
        # That's a fatal error

        status, code = 'unknown_status', 'unknown_error_code'
        if 'response' in data:
            status = data['response'].get('status', status)
            code = data['response'].get('error_code', code)
        if not title:
            title = "BACKEND ERROR"
        if not msg:
            msg = "%s: %s %s %s" % (title, status, code, ApiPool().current_server_name)
    else:
        # That's a non-fatal error

        message = data
        log.error(message)

        data = {}
        data['message'] = message

        # Formatting traceback may raise a UnicodeDecodeError...
        data['stack'] = []
        try:
            data['stack'] = [l for l in traceback.format_stack()]
        except Exception as ee:
            pass

        populate_error_report(data)
        if caught:
            data['error_caught'] = "%s" % caught

        # inspect may raise a UnicodeDecodeError...
        fname = ''
        try:
            fname = inspect.stack()[1][3]
        except Exception as e:
            pass

        if title:
            msg = title
        else:
            title = "NON-FATAL BACKEND ERROR"
            msg = "%s %s %s(): %s" % (title, ApiPool().current_server_name, fname, message)

    if os.environ.get('NO_ERROR_REPORTING', '') == '1':
        log.info("NO_ERROR_REPORTING is set: not sending error to slack or email")
        return

    global crash_reporter
    log.info("Reporting crash...")

    try:
        crash_reporter(msg, json.dumps(data, sort_keys=True, indent=4))
    except Exception as e:
        # Don't block on replying to api caller
        log.error("Failed to send email report: %s" % str(e))


def crash_handler(f):
    """Return a decorator that reports failed api calls via the error_reporter,
    for use on every server endpoint"""

    @wraps(f)
    def wrapper(*args, **kwargs):
        """Generate a report of this api call, and if the call failed or was too slow,
        forward this report via the error_reporter"""

        data = {}
        t0 = timenow()

        # Call endpoint and log execution time
        try:
            res = f(*args, **kwargs)
        except Exception as e:
            # An unhandled exception occured. Forge a Response
            exc_type, exc_value, exc_traceback = sys.exc_info()
            s = ''.join((traceback.format_exception(exc_type, exc_value, exc_traceback, 30)))
            e = UnhandledServerError(s)
            log.error("UNHANDLED EXCEPTION: %s" % s)
            res = e.http_reply()

        t1 = timenow()

        # Is the response an Error instance?
        response_type = type(res).__name__
        status_code = ''
        is_an_error = 0
        error = ''
        error_description = ''
        error_user_message = ''

        error_id = ''

        if isinstance(res, Response):
            # Got a flask.Response object
            status_code = res.status_code
            if (str(status_code) != '200'):
                # Assuming it is a PntCommonError.http_reply()

                res_data = res.get_data()

                if type(res_data) is bytes:
                    res_data = res_data.decode("utf-8")

                is_json = True
                try:
                    j = json.loads(res_data)
                except ValueError as e:
                    # This was a plain html response. Fake an error
                    is_json = False
                    j = {'error': res_data}

                error = j.get('error', 'NO_ERROR_IN_JSON')
                error_description = j.get('error_description', res_data)
                if error_description == '':
                    error_description = res_data

                error_user_message = j.get('user_message', '')
                is_an_error = 1

                # Patch Response to contain a unique id
                if is_json and 'error_id' not in j:
                    # If the error is forwarded by multiple micro-services, we
                    # want the error_id to be set only on the original error
                    error_id = str(uuid.uuid4())
                    j['error_id'] = error_id
                    res.set_data(json.dumps(j))
            else:
                status_code = 200

            request_args = []
            if len(args):
                request_args.append(args)
            if kwargs:
                request_args.append(kwargs)

            data = {
                # Set only on the original error, not on forwarded ones, not on
                # success responses
                'error_id': error_id,

                # Call results
                'time': {
                    'start': t0.isoformat(),
                    'end': t1.isoformat(),
                    'microsecs': (t1 - t0).microseconds,
                },

                # Response details
                'response': {
                    'type': response_type,
                    'status': status_code,
                    'is_error': is_an_error,
                    'error_code': error,
                    'error_description': error_description,
                    'user_message': error_user_message,
                },

                # Request details
                'request': {
                    'params': pformat(request_args),
                },
            }

            # Did klue-client-server set a call_id and call_path?
            call_id, call_path = '', ''
            if hasattr(stack.top, 'call_id'):
                call_id = stack.top.call_id
            if hasattr(stack.top, 'call_path'):
                call_path = stack.top.call_path

        # Unique ID associated to all responses associated to a given
        # call to klue-api, across all micro-services
        data['call_id'] = call_id
        data['call_path'] = call_path

        # If user is authenticated, get her id
        user_data = {
            'id': '',
            'is_auth': 0,
            'ip': request.remote_addr,
        }

        if 'X-Forwarded-For' in request.headers:
            user_data['forwarded_ip'] = request.headers.get('X-Forwarded-For', '')

        if 'User-Agent' in request.headers:
            user_data['user_agent'] = request.headers.get('User-Agent', '')

        if hasattr(stack.top, 'current_user'):
            user_data['is_auth'] = 1
            user_data['id'] = stack.top.current_user.get('sub', '')
            for k in ('name', 'email', 'is_expert', 'is_admin', 'is_support', 'is_tester', 'language'):
                v = stack.top.current_user.get(k, None)
                log.info("Got %s=%s" % (k, v))
                if v:
                    user_data[k] = v

        data['user'] = user_data

        # Is the current code running as a server?
        if ApiPool().current_server_api:
            # Server info
            server = request.base_url
            server = server.replace('http://', '')
            server = server.replace('https://', '')
            server = server.split('/')[0]
            parts = server.split(':')
            fqdn = parts[0]
            port = parts[1] if len(parts) == 2 else ''

            data['server'] = {
                'fqdn': fqdn,
                'port': port,
                'api_name': ApiPool().current_server_name,
                'api_version': ApiPool().current_server_api.get_version(),
            }

            # Endpoint data
            data['endpoint'] = {
                'id': "%s %s %s" % (ApiPool().current_server_name, request.method, request.path),
                'url': request.url,
                'base_url': request.base_url,
                'path': request.path,
                'method': request.method
            }

            log.info("Analytics: " + pformat(data))

        #
        # Should we report this call?
        #

        fqdn = data['server']['fqdn']
        if 'server' not in data:
            log.info("This is a unitest running: not reporting error")
            pass
        elif '192.168' in fqdn or '127.0.0' in fqdn:
            log.info("This is a test setup: not reporting error")
            pass
        else:
            # If it is an internal errors, report it
            # If it is too slow, report it
            if str(data['response']['status']) in ('500'):
                report_error(data)
            elif int(data['time']['microsecs']) > 3000000:
                report_error(data, msg="SLOW BACKEND CALL: %s %s" % (ApiPool().current_server_name, data['endpoint']['path']))

        return res

    return wrapper



import logging
import sys

DEFAULT_LEVEL = logging.DEBUG

root = logging.getLogger()
ch = logging.StreamHandler(sys.stdout)

formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
root.addHandler(ch)
root.setLevel(DEFAULT_LEVEL)

# Make an exception for boto and bravado: its debug level is just too verbose...
logging.getLogger('boto').setLevel(logging.INFO)
logging.getLogger('bravado_core.spec').setLevel(logging.INFO)
logging.getLogger('swagger_spec_validator.ref_validators').setLevel(logging.INFO)


def get_logger():
    global root
    return root

def set_level(newlevel):
    global root
    root.setLevel(newlevel)



import pprint
import jwt
import base64
import logging
from urllib.parse import unquote_plus
from contextlib import contextmanager
from functools import wraps
from flask import request
from klue_microservice.exceptions import AuthInvalidTokenError, AuthInvalidTokenError
from klue_microservice.exceptions import AuthMissingHeaderError, KlueMicroServiceException
from klue_microservice.utils import timenow, to_epoch

try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack


log = logging.getLogger(__name__)


#
# Configuration
#

DEFAULT_JWT_ISSUER = 'klue'
DEFAULT_JWT_AUDIENCE = 'HFhAcAZ1VdRt0anWpefDfGYnW8F79uLF'
DEFAULT_USER_ID = None
DEFAULT_TOKEN_TIMEOUT = 86400
# Automatically renew token if expires in less than this time (in sec) 3 hours
# in sec
DEFAULT_TOKEN_RENEW_AFTER = 10800
DEFAULT_JWT_SECRET = None

def set_jwt_defaults(issuer=None, user_id=None, token_timeout=None, token_renew=None, secret=None, audience=None):
    if issuer:
        global DEFAULT_JWT_ISSUER
        log.info("Setting JWT issuer to %s" % issuer)
        DEFAULT_JWT_ISSUER = issuer
    if user_id:
        global DEFAULT_USER_ID
        log.info("Setting JWT user_id to %s" % user_id)
        DEFAULT_USER_ID = user_id
    if token_timeout:
        global DEFAULT_TOKEN_TIMEOUT
        log.info("Setting JWT timeout to %s" % token_timeout)
        DEFAULT_TOKEN_TIMEOUT = token_timeout
    if token_renew:
        global DEFAULT_TOKEN_RENEW_AFTER
        log.info("Setting JWT renew timeout to %s" % token_renew)
        DEFAULT_TOKEN_RENEW_AFTER = token_renew
    if secret:
        global DEFAULT_JWT_SECRET
        log.info("Setting JWT secret %s*****.." % secret[0:5])
        DEFAULT_JWT_SECRET = secret
    if audience:
        global DEFAULT_JWT_AUDIENCE
        log.info("Setting JWT audience: %s" % audience)
        DEFAULT_JWT_AUDIENCE = audience


#
# Decorators used to add authentication to endpoints in swagger specs
#

def requires_auth(f):
    """A decorator for flask api methods that validates auth0 tokens, hence ensuring
    that the user is authenticated. Code coped from:
    https://github.com/auth0/auth0-python/tree/master/examples/flask-api
    """

    @wraps(f)
    def requires_auth_decorator(*args, **kwargs):

        try:
            authenticate_http_request()
        except KlueMicroServiceException as e:
            return e.http_reply()

        return f(*args, **kwargs)

    return requires_auth_decorator


def add_auth(f):
    """A decorator that adds the authentication header to requests arguments"""

    def add_auth_decorator(*args, **kwargs):
        token = get_user_token()
        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers']['Authorization'] = "Bearer %s" % token
        return f(*args, **kwargs)

    return add_auth_decorator


#
# Get and validate a token
#

def load_auth_token(token, load=True):
    """Validate an auth0 token. Returns the token's payload, or an exception
    of the type:"""

    assert DEFAULT_JWT_SECRET

    log.info("Extracting issuer of token")

    # First extract the issuer (default to 'klue')
    issuer = DEFAULT_JWT_ISSUER
    try:
        headers = jwt.get_unverified_header(token)
    except jwt.DecodeError:
        raise AuthInvalidTokenError('token signature is invalid')

    log.debug("Token has headers %s" % headers)

    if 'iss' in headers:
        issuer = headers['iss']

    # Then validate the token against this issuer
    log.info("Validating token in issuer %s" % issuer)
    try:
        payload = jwt.decode(
            token,
            DEFAULT_JWT_SECRET,
            audience=DEFAULT_JWT_AUDIENCE,
            # Allow for a time difference of up to 5min (300sec)
            leeway=300
        )
    except jwt.ExpiredSignature:
        raise AuthTokenExpiredError('Auth token is expired')
    except jwt.InvalidAudienceError:
        raise AuthInvalidTokenError('incorrect audience')
    except jwt.DecodeError:
        raise AuthInvalidTokenError('token signature is invalid')
    except jwt.InvalidIssuedAtError:
        raise AuthInvalidTokenError('Token was issued in the future')

    # Save payload to stack
    payload['token'] = token
    payload['iss'] = issuer

    if load:
        stack.top.current_user = payload

    return payload

def authenticate_http_request():
    """Validate auth0 tokens passed in the request's header, hence ensuring
    that the user is authenticated. Code copied from:
    https://github.com/auth0/auth0-python/tree/master/examples/flask-api

    Return a PntCommonException if failed to validate authentication.
    Otherwise, return the token's payload (Also stored in stack.top.current_user)
    """


    auth = request.headers.get('Authorization', None)

    if not auth:
        auth = request.cookies.get('token', None)
        if auth:
            auth = unquote_plus(auth)

    log.debug("Validating Auth header [%s]" % auth)

    if not auth:
        raise AuthMissingHeaderError('There is no Authorization header in the HTTP request')

    parts = auth.split()

    if parts[0].lower() != 'bearer':
        raise AuthInvalidTokenError('Authorization header must start with Bearer')
    elif len(parts) == 1:
        raise AuthInvalidTokenError('Token not found in Authorization header')
    elif len(parts) > 2:
        raise AuthInvalidTokenError('Authorization header must be Bearer + \s + token')

    token = parts[1]

    return load_auth_token(token)

#
# Generate tokens
#

def generate_token(user_id, expire_in=None, data={}, issuer=DEFAULT_JWT_ISSUER, iat=None):
    """Generate a new JWT token for this user_id. Default expiration date
    is 1 year from creation time"""
    assert user_id
    assert isinstance(data, dict)
    assert DEFAULT_JWT_SECRET

    if expire_in is None:
        expire_in = DEFAULT_TOKEN_TIMEOUT

    if iat:
        epoch_now = iat
    else:
        epoch_now = to_epoch(timenow())
    epoch_end = epoch_now + expire_in

    data['iss'] = issuer
    data['sub'] = user_id
    data['aud'] = DEFAULT_JWT_AUDIENCE
    data['exp'] = epoch_end
    data['iat'] = epoch_now

    headers = {
        "typ": "JWT",
        "alg": "HS256",
        "iss": issuer
    }

    log.debug("Encoding token with data %s and headers %s" % (data, headers))

    t = jwt.encode(
        data,
        DEFAULT_JWT_SECRET,
        headers=headers,
    )

    if type(t) is bytes:
        t = t.decode("utf-8")

    return t


@contextmanager
def backend_token(issuer=DEFAULT_JWT_ISSUER, user_id=DEFAULT_USER_ID):

    assert issuer
    assert user_id

    cur_token = ''
    if not hasattr(stack.top, 'current_user'):
        stack.top.current_user = {}
    else:
        cur_token = stack.top.current_user.get('token', '')

    tmp_token = generate_token(user_id, issuer=issuer)

    log.debug("Temporarily using custom token for %s and issuer %s: %s" % (user_id, issuer, tmp_token))
    stack.top.current_user['token'] = tmp_token
    yield tmp_token
    log.debug("Restoring token %s" % cur_token)
    stack.top.current_user['token'] = cur_token


#
# Access token data during runtime
#


def get_userid():
    """Return the authenticated user's id, i.e. its auth0 id"""
    current_user = stack.top.current_user
    return current_user.get('sub', '')


def get_user_token():
    """Return the authenticated user's auth token"""
    current_user = stack.top.current_user
    return current_user.get('token', '')


def get_token_issuer():
    """Return the issuer in which this user's token was created"""
    try:
        current_user = stack.top.current_user
        return current_user.get('iss', get_config().default_jwt_issuer)
    except Exception:
        pass
    return get_config().default_jwt_issuer



import os
import sys
import logging
import datetime
from dateutil import parser
import socket
import logging
import pytz


log = logging.getLogger(__name__)


def is_ec2_instance():
    """Try fetching instance metadata at 'curl http://169.254.169.254/latest/meta-data/'
    to see if host is on an ec2 instance"""

    # Note: this code assumes that docker containers running on ec2 instances
    # inherit instances metadata, which they do as of 2016-08-25

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(0.2)
    try:
        s.connect(("169.254.169.254", 80))
        return True
    except socket.timeout:
        return False
    except socket.error:
        return False


def timenow():
    return datetime.datetime.now(pytz.timezone('utc'))


def to_epoch(t):
    """Take a datetime, either as a string or a datetime.datetime object,
    and return the corresponding epoch"""
    if isinstance(t, str):
        if '+' not in t:
            t = t + '+00:00'
        t = parser.parse(t)
    elif t.tzinfo is None or t.tzinfo.utcoffset(t) is None:
        t = t.replace(tzinfo=pytz.timezone('utc'))

    t0 = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, pytz.timezone('utc'))
    delta = t - t0
    return int(delta.total_seconds())


def to_datetime(e):
    """Take an epoch and return a timezone aware datetime"""
    return datetime.datetime.fromtimestamp(e, pytz.timezone('utc'))


def get_container_version():
    """Return the version of the docker container running the present server,
    or '' if not in a container"""
    root_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
    version_file = os.path.join(root_dir, 'VERSION')
    if os.path.exists(version_file):
        with open(version_file) as f:
            return f.read()
    return ''
